!
! Simple program to implement a broadcast using point-to-point
! communications. Although this is a very useful teaching exercise
! which illustrates the communications pattern, in a real MPI program
! you should call the collective routine MPI_Bcast().
!

program scatter

  use mpi

  implicit none

  integer :: i, src, dest, rank, size, root, tag, count, ierr
  integer :: comm
  integer, dimension(MPI_STATUS_SIZE) :: status

  integer, parameter :: N = 12

  integer, dimension(N) :: x

  comm = MPI_COMM_WORLD
  root = 0   ! The process that contains the data to be broadcast

  call MPI_Init(ierr)

  call MPI_Comm_size(comm, size, ierr)
  call MPI_Comm_rank(comm, rank, ierr)

  if (rank == 0) write(*,*) "Running broadcast program on ", size, " processes"

  do i = 1, N

     if (rank == root) then
        x(i) = i-1
     else
        x(i) = -1
     end if

  end do

  write(*,*) "On rank ", rank, " x = ", x(:)

  count = N/size;

  tag = 0

  if (rank == root) then

     do dest = 0, size-1

        if (dest /= root) call MPI_Ssend(x(dest*count+1), count, MPI_INTEGER, dest, tag, comm, ierr)

     end do

  else

     call MPI_Recv(x, count, MPI_INTEGER, root, tag, comm, status, ierr)

  end if

  write(*,*) "On rank ", rank, " x = ", x(:)

  call MPI_Finalize(ierr)

end program scatter

