!
! Perform a global sum of data stored on each process by rotating each
! piece of data all the way round the ring. At each iteration, a
! process receives some data from the left, adds the value to its
! running total, then passes the data it has just received on to the
! right.
!

program rotate

  use mpi

  implicit none

  integer :: comm, ierr, rank, size, request,tag
  integer :: send_status(MPI_STATUS_SIZE), recv_status(MPI_STATUS_SIZE)
  integer :: left, right, addon, sum, i, passon

  comm = MPI_COMM_WORLD
  tag = 1

  call MPI_Init(ierr)
  call MPI_Comm_rank(comm,rank,ierr)
  call MPI_Comm_size(comm,size,ierr)

  ! Send current number to the right and receive from the left
  left = rank - 1
  right = rank + 1

  if (rank.eq.size-1) then
     right = 0
  endif

  if(rank.eq.0) then
     left = size -1
  endif

  sum = 0
  ! Initialise local values to:
  passon = rank
  ! Alternatively use
  ! passon = (rank+1)**2

  ! Use non-blocking point-to-point communication 

  do i=1,size

     call MPI_Issend(passon,1,MPI_INTEGER,right,tag,comm,request,ierr)
     call MPI_Recv(addon,1,MPI_INTEGER,left,tag,comm,recv_status,ierr)
     call MPI_Wait(request,send_status,ierr)

     sum = sum + addon
     passon = addon

  enddo

  write(*,*) "The sum is: ", sum," on rank ", rank

  call MPI_Finalize(ierr)

end program rotate
